# -*- coding: utf-8 -*-
"""
Created on Thu Mar 29 14:04:58 2018

Pyene Networks provides methods to simulate power networks and generation units
(i.e., Unit commitment)

@author: Dr Eduardo Alejandro Martínez Ceseña
https://www.researchgate.net/profile/Eduardo_Alejandro_Martinez_Cesena
"""
from __future__ import division
from pyomo.core import Constraint, Var, NonNegativeReals, Reals, Binary
import numpy as np
import json
import warnings


class pyeneNConfig:
    ''' Default settings used for this class '''
    def __init__(self):
        # Basic settings
        self.settings = {
                'File': None,  # File to be loaded
                'Flag': True,  # Add electricity network
                'NoTime': 1,  # Number of time steps
                'SecurityFlag': False,  # Enable all security constraints
                'Security': [],  # Security constraints (lines)
                'Losses': True,  # Consideration of losses
                'Feasibility': True,  # Feasibility constraints (curtailment)
                'Load_Type': [],  # 0 Urban, 1 Rural
                'Pieces': [],  # Size of pieces (MW) for piece-wise estimations
                'Loss': None,  # Factor for losses
                'Ancillary': None,  # Need for uncillary services
                'UC': False,  # Model UC
                }
        # Connections
        self.connections = {
                'set': range(1),  # Connections between nodes
                'Branches': 0,  # Real number of branches (including parallel)
                'Flow': [0],  # Power flow through the lines
                'Voltage': [0],  # Voltages in each node
                'Loss': [0],  # Power losses
                'Loss_Param': [0, 0.01, 0.025, 0.05, 0.1, 0.2, 0.4, 0.8, 1.2,
                               1.6, 2, 2.5, 3],  # To model power losses
                'Feasibility': [0],  # Lines to trip for security consideration
                'Generation': [0],  # Location (bus) of generators
                'Cost': [0],  # Generator consts
                'Pump': [0]  # Pumps
                }
        # Scenarios
        self.scenarios = {
                'Number': 1,  # Number of scenarios
                'Links': 'Default',  # Links between the buses and profiles
                'NoDem': 0,  # Number of demand profiles
                'Demand': [1],  # Demand profiles
                'NoRES': 0,  # Number of RES profiles
                'LinksRes': 'Default',  # Links RES generators and profiles
                'RES': [],  # Location of the RES profiles
                'Weights': None  # Weight of the time period
                }
        # Conventional generators
        self.conventional = {
                'Number': None,  # Number of conventional generators
                'Ancillary': [True],  # Can it provide ancillary services?
                'Baseload': [0],  # 0-1 for the use of conv for baseload
                'Ramp': [],  # Set ramps for conventional generators
                'RES': [True],  # Can it support RES integration?
                'MUT': [],  # MInimum up time
                'MDT': []  # Minimum down time
                }
        # Hydropower
        self.hydropower = {
                'Number': 0,  # Number of hydropower plants
                'Bus': [],  # Location (Bus) in the network
                'Max': [],  # Capacity (MW)
                'Min': [],  # Minimum generation
                'Cost': [],  # Costs
                'Ramp': [None],  # Ramp
                'Baseload': [0],  # 0-1 for the use of water for baseload
                'Ancillary': [True],  # Can it provide ancillary services?
                'RES': [True],  # Can it support RES integration?
                }
        # Pumps
        self.pumps = {
                'Number': 0,  # Number of pumps
                'Bus': [],  # Location (Bus) in the network
                'Max': [],  # Capacity (kW)
                'Value': []  # Value (OF)
                }
        # RES
        self.RES = {
                'Number': 0,  # Number of RES generators
                'Bus': [],  # Location (Bus) in the network
                'Max': [],  # Capacity (kW)
                'Cost': [],  # Cost (OF)
                'Uncertainty': [None]  # Introduce reserve needs
                }
        self.Storage = {
                'Number': 0,  # Number of storage units
                'Bus': [],  # Location (Bus) in the network
                'Max': [],  # Capacity (kW)
                'Efficiency': []  # Round trip efficiency
                }
        self.Print = {
                'Generation': True,
                'Flows': True,
                'Voltages': True,
                'Losses': True,
                'Curtailment': True,
                'Pump': True,
                'Services': True,
                'GenBus': True,
                'UC': True,
                }
        # TODO: Remove reference
        self.Aux = {
                'shift': 1
                }


class ENetworkClass:
    def __init__(self, obj=None):
        ''' Initialise network class '''
        # Get default values
        if obj is None:
            obj = pyeneNConfig()

        # Copy attributes
        for pars in obj.__dict__.keys():
            setattr(self, pars, getattr(obj, pars))

        # sets and parameters used for the mathematical model
        self.s = {}
        self.p = {}

    def _getLL(self, NoLL, NoDt, DtLL):
        ''' Produce LL while removing the 'next' '''
        # Produce LL and 'next'
        LL = np.zeros(NoLL, dtype=int)
        LLnext = np.zeros(NoDt, dtype=int)
        for xd in range(NoDt):
            xpos = DtLL[xd]
            # Is it the first one?
            if LL[xpos-1] == 0:
                LL[xpos-1] = xd+1
            else:
                xpos = LL[xpos-1]
                while LLnext[xpos-1] != 0:
                    xpos = LLnext[xpos-1]
                LLnext[xpos-1] = xd+1

        # Sort data so that the 'next' can be removed
        LL1 = np.zeros(NoDt+1, dtype=int)
        LL2 = np.zeros((NoLL, 2), dtype=int)
        xL = 1
        for xn in range(NoLL):
            xpos = LL[xn]
            if xpos != 0:
                LL2[xn][0] = xL
                # Add generators
                while xpos != 0:
                    LL1[xL] = xpos
                    xpos = LLnext[xpos-1]
                    xL += 1
                LL2[xn][1] = xL-1

        return (LL1, LL2)

    def addCon(self, m):
        ''' Add pyomo constraints '''
        # Currently using the same feasibility variable as the generators
        # TODO: Assign dedicated curtailment/spilling variable
        if self.scenarios['NoDem'] == 0:
            self.p['daux'] = 0
        else:
            self.p['daux'] = 1
        if self.settings['Feasibility']:
            self.p['faux'] = 1
        else:
            self.p['faux'] = 0

        # Is the network enabled
        if self.settings['Flag']:
            # Branch flows
            m.cNEFlow = Constraint(self.s['Tim'], self.s['Bra'],
                                   self.s['Sec2'], self.s['Con'],
                                   rule=self.cNEFlow_rule)
            # Branch capacity (Positive)
            m.cNEFMax = Constraint(self.s['Tim'], self.s['Bra'],
                                   self.s['Sec2'], self.s['Con'],
                                   rule=self.cNEFMax_rule)
            # Branch capacity (Negative)
            m.cNEFMin = Constraint(self.s['Tim'], self.s['Bra'],
                                   self.s['Sec2'], self.s['Con'],
                                   rule=self.cNEFMin_rule)

            if self.settings['Losses']:
                m.cNDCLossA = Constraint(self.s['Bra'], self.s['Loss'],
                                         self.s['Tim'], self.s['Con'],
                                         rule=self.cNDCLossA_rule)
                m.cNDCLossB = Constraint(self.s['Bra'], self.s['Loss'],
                                         self.s['Tim'], self.s['Con'],
                                         rule=self.cNDCLossB_rule)

            # Balance: Gen + Flow in - loss/2 = Demand + flow out + loss/2
            m.cNEBalance = Constraint(self.s['Bus'], self.s['Tim'],
                                      self.s['Sec2'], self.s['Con'],
                                      rule=self.cNEBalance_rule)
        else:
            # Create additional paremeters
            # Addition of power losses
            if self.settings['Loss'] is None:
                self.p['LossM'] = 1
            else:
                self.p['LossM'] = 1+self.settings['Loss']

            # Balance: Gen = Demand
            m.cNEBalance0 = Constraint(self.s['Tim'], self.s['Sec2'],
                                       self.s['Con'],
                                       rule=self.cNEBalance0_rule)
        # Is there baseload
        if sum(self.hydropower['Baseload']) > 0:
            self.scenarios['WSum'] = sum(self.scenarios['Weights'])
            m.cNEBaseload = Constraint(self.s['Gen'], self.s['Tim'],
                                       self.s['Con'],
                                       rule=self.cNEBaseload_rule)
        if self.settings['UC']:
            # Maximum generation
            m.cNEGMax = Constraint(self.s['Gen'], self.s['Tim'], self.s['Con'],
                                   rule=self.cNEGMaxUC_rule)
            # Minimum generation
            m.cNEGMin = Constraint(self.s['Gen'], self.s['Tim'], self.s['Con'],
                                   rule=self.cNEGMinUC_rule)
            self.LLUC = np.zeros(self.settings['NoTime']*2, dtype=int)
            x1 = 1
            for x in range(self.settings['NoTime']*2):
                self.LLUC[x] = x1
                x1 += 1
                if x1 == self.settings['NoTime']:
                    x1 = 0

            # Minimum down time
            m.cNEGMinDT1 = Constraint(self.s['Gen'], self.s['Tim'],
                                      self.s['Con'], rule=self.cNEGMinDT1_rule)
            # Minimum up time
            m.cNEGMinUT1 = Constraint(self.s['Gen'], self.s['Tim'],
                                      self.s['Con'], rule=self.cNEGMinUT1_rule)
            # Minimum down time
            m.cNEGMinDT2 = Constraint(self.s['Gen'], self.s['Tim'],
                                      self.s['Con'], rule=self.cNEGMinDT2_rule)
            # Minimum up time
            m.cNEGMinUT2 = Constraint(self.s['Gen'], self.s['Tim'],
                                      self.s['Con'], rule=self.cNEGMinUT2_rule)

        else:
            # Maximum generation
            m.cNEGMax = Constraint(self.s['Gen'], self.s['Tim'], self.s['Con'],
                                   rule=self.cNEGMax_rule)
            # Minimum generation
            m.cNEGMin = Constraint(self.s['Gen'], self.s['Tim'], self.s['Con'],
                                   rule=self.cNEGMin_rule)

        # Piece-wise generation costs approximation
        m.cNEGenC = Constraint(self.s['Gen'], range(self.Gen.get_NoPieces()),
                               self.s['Tim'], self.s['Con'],
                               rule=self.cNEGenC_rule)

        # Dinamic load (pump) maximum capacity
        m.cNDLMax = Constraint(self.s['Pump'], self.s['Tim'], self.s['Con'],
                               rule=self.cNLDMax_rule)
        # Dinamic load (pump) initialisation
        m.cNDLIni = Constraint(self.s['Tim'], self.s['Con'],
                               rule=self.cNLDIni_rule)

        # Adding RES limits
        if self.RES['Number'] > 0:
            m.cNRESMax = Constraint(self.s['Tim'], self.s['RES'],
                                    self.s['Con'], rule=self.cNRESMax_rule)
        # Storage
        if self.Storage['Number'] > 0:
            m.cNStoreMax = Constraint(m.sNSto, self.s['Tim'], self.s['Con'],
                                      rule=self.cNStoreMax_rule)
            m.cNStoreMin = Constraint(m.sNSto, self.s['Tim'], self.s['Con'],
                                      rule=self.cNStoreMin_rule)
        m.cNStore0 = Constraint(self.s['Tim'], rule=self.cNStore0_rule)

        # Adding generation ramps
        m.cNGenRampUp = Constraint(self.s['Gen'], self.s['Tim'], self.s['Con'],
                                   rule=self.cNGenRampUp_rule)
        m.cNGenRampDown = Constraint(self.s['Gen'], self.s['Tim'],
                                     self.s['Con'],
                                     rule=self.cNGenRampDown_rule)

        # Adding service constraints
        if len(self.s['GServices']) > 0:
            m.cNServices = Constraint(self.s['Tim'], self.s['Con'],
                                      rule=self.cNServices_rule)
            if len(self.s['GAncillary']) > 0:
                m.cNServicesA = Constraint(self.s['Tim'], self.s['Con'],
                                           rule=self.cNServicesA_rule)
                m.cNAncillary = Constraint(self.s['Tim'], self.s['Con'],
                                           rule=self.cNAncillary_rule)

            if len(self.s['GRES']) > 0:
                m.cNServicesR = Constraint(self.s['Tim'], self.s['Con'],
                                           rule=self.cNServicesR_rule)
                m.cNUncRES = Constraint(self.s['Tim'], self.s['Con'],
                                        rule=self.cNUncRES_rule)

        if len(self.s['GRES']) == 0 and self.RES['Number'] > 0 and \
                self.RES['Uncertainty'][0] is not None:
            m.cNUncRES0 = Constraint(self.s['Tim'], self.s['Con'],
                                     rule=self.cNUncRES0_rule)

        return m

    def addPar(self, m):
        ''' Add pyomo parameters '''
        self.p['MaxPump'] = self.pumps['Max']

        return m

    def addSets(self, m):
        ''' Add pyomo sets '''
        self.s['Con'] = self.connections['set']
        self.s['Bra'] = range(self.connections['Branches'])
        self.s['Bus'] = range(self.ENetwork.get_NoBus())
        self.s['Buses'] = range(self.NoBuses)
        self.s['Pump'] = range(self.pumps['Number'])
        self.s['Fea'] = range(self.NoFea)
        self.s['Gen'] = range(self.Gen.get_NoGen())
        self.s['Tim'] = range(self.settings['NoTime'])
        self.s['Sec1'] = range(self.NoSec1)
        self.s['Sec2'] = range(self.NoSec2+1)
        self.s['Sto'] = range(self.Storage['Number'])
        self.s['RES'] = range(self.RES['Number'])

        if self.settings['Losses']:
            self.s['Loss'] = range(self.Number_LossCon)

        return m

    def addVars(self, m):
        ''' Add pyomo variables '''
        Noh = len(self.s['Con'])
        # Is the network enabled
        if self.settings['Flag']:
            m.vNFlow = Var(range(Noh*(self.NoBranch)), self.s['Tim'],
                           domain=Reals, initialize=0.0)
            m.vNVolt = Var(range(Noh*(self.NoBuses)), self.s['Tim'],
                           domain=Reals, initialize=0.0)
            if self.settings['Losses']:
                m.vNLoss = Var(range(Noh*self.connections['Branches']),
                               self.s['Tim'], domain=NonNegativeReals,
                               initialize=0.0)

        m.vNPump = Var(range(Noh*(self.pumps['Number']+1)), self.s['Tim'],
                       domain=NonNegativeReals, initialize=0.0)
        m.vNFea = Var(range(Noh*self.NoFea), self.s['Tim'],
                      domain=NonNegativeReals, initialize=0.0)
        m.vNGCost = Var(range(Noh*self.Gen.get_NoGen()), self.s['Tim'],
                        domain=NonNegativeReals, initialize=0.0)
        m.vNGen = Var(range(Noh*(self.Gen.get_NoGen()+1)), self.s['Tim'],
                      domain=NonNegativeReals, initialize=0.0)
        m.vNStore = Var(range(Noh*(self.Storage['Number'])+1), self.s['Tim'],
                        domain=NonNegativeReals, initialize=0.0)
        if len(self.s['GServices']) > 0:
            m.vNServ = Var(range(Noh*self.p['GServices']), self.s['Tim'],
                           domain=NonNegativeReals, initialize=0.0)
        # Add Unit commitment constraints
        if self.settings['UC']:
            m.vNGen_Bin = \
                Var(range(Noh*(self.Gen.get_NoBin()+1)), self.s['Tim'],
                    domain=Binary, initialize=1.0)
            m.vNGen_MUT = \
                Var(range(Noh*(self.Gen.get_NoMUT()+1)), self.s['Tim'],
                    domain=NonNegativeReals, bounds=(0.0, 1.0), initialize=0.0)
            m.vNGen_MDT = \
                Var(range(Noh*(self.Gen.get_NoMDT()+1)), self.s['Tim'],
                    domain=NonNegativeReals, bounds=(0.0, 1.0), initialize=0.0)

        return m

    def cNAncillary_rule(self, m, xt, xh):
        ''' Ancillary services constraint '''
        return m.vNServ[xh*self.p['GServices'], xt] >= \
            self.settings['Ancillary'] * \
            sum(m.vNGen[self.connections['Generation'][xh]+x, xt]
                for x in self.Gen.get_GenAll()) - \
            sum(m.vNFea[xh, xt] for xf in range(self.p['faux']))

    def cNDCLossA_rule(self, m, xb, xL, xt, xh):
        ''' Power losses (Positive) '''
        return self.ENetwork.cNDCLossA_rule(m, xt, xb, xL,
                                            self.connections['Flow'][xh],
                                            self.connections['Loss'][xh])

    def cNDCLossB_rule(self, m, xb, xL, xt, xh):
        ''' Power losses (Negative) '''
        return self.ENetwork.cNDCLossB_rule(m, xt, xb, xL,
                                            self.connections['Flow'][xh],
                                            self.connections['Loss'][xh])

    def cNEBalance_rule(self, m, xn, xt, xs, xh):
        ''' Nodal balance:
        Generation + Flow in - loss/2 = Demand + flow out + loss/2
        '''
        # Check for case without demand profiles
        if self.LLStor[xn, xh] == 0:
            aux = 0
        else:
            aux = self.Storage['Efficiency'][self.LLStor[xn, 0]-1]

        # Get losses
        if self.settings['Losses']:
            # The model estimates the losses
            LossF = sum(m.vNLoss[self.connections['Loss'][xh]+x1, xt]/2
                        for x1 in self.ENetwork.Bus[xn].get_LossF())
            LossT = sum(m.vNLoss[self.connections['Loss'][xh]+x2, xt]/2
                        for x2 in self.ENetwork.Bus[xn].get_LossT())
            LossM = 1
        else:
            # The losses are pre-calculated
            LossF = 0.5 * \
                sum(self.ENetwork.Branch[xb].getLoss() for xb in
                    self.ENetwork.get_FlowF(xn, xs))/self.ENetwork.get_Base()
            LossT = 0.5 * \
                sum(self.ENetwork.Branch[xb].getLoss() for xb in
                    self.ENetwork.get_FlowT(xn, xs))/self.ENetwork.get_Base()

            if self.settings['Loss'] is None:
                LossM = 1
            else:
                LossM = 1+self.settings['Loss']
                

        return (sum(m.vNGen[self.connections['Generation'][xh]+xg, xt]
                    for xg in self.Gen.get_GenInBus(self.ENetwork.Bus[xn])) +
                sum(m.vNFlow[self.connections['Flow'][xh]+x2, xt]
                    for x2 in self.ENetwork.get_FlowT(xn, xs))-LossT ==
                LossM*self.busData[xn]*self.scenarios['Demand']
                                               [xt*self.p['daux'] +
                                                self.busScenario[xn][xh]] -
                (m.vNStore[self.LLStor[xn, xh], xt] -
                 m.vNStore[self.LLStor[xn, xh], self.LLTime[xt]])*aux /
                self.scenarios['Weights'][xt] -
                sum(m.vNFea[self.connections['Feasibility'][xh] +
                            self.p['LLFea2'][xn], xt] for x1 in
                    range(self.p['LLFea1'][xn])) +
                m.vNPump[self.connections['Pump'][xh]+self.p['LLPump'][xn],
                         xt] +
                sum(m.vNFlow[self.connections['Flow'][xh]+x1, xt]
                    for x1 in self.ENetwork.get_FlowF(xn, xs)) +
                LossF)

    def cNEBalance0_rule(self, m, xt, xs, xh):
        ''' Nodal balance without networks '''
        return float(sum(self.busData[xn]*self.scenarios['Demand']
                     [xt*self.p['daux']+self.busScenario[xn][xh]]
                     for xn in self.s['Bus']))*self.p['LossM'] + \
            + sum(m.vNPump[xh*(self.pumps['Number']+1)+xp+1, xt]
                  for xp in self.s['Pump']) == \
            sum(m.vNFea[xh, xt] for xf in range(self.p['faux'])) + \
            sum(m.vNGen[self.connections['Generation'][xh]+xg, xt]
                for xg in self.Gen.get_GenAll()) - \
            sum((m.vNStore[xh*(self.Storage['Number']+1)+xn+1, xt] -
                 m.vNStore[xh*(self.Storage['Number']+1)+xn+1,
                           self.LLTime[xt]])*self.Storage['Efficiency'][xn] /
                self.scenarios['Weights'][xt] for xn in self.s['Sto'])

    def cNEBaseload_rule(self, m, xg, xt, xh):
        ''' Baseload rule '''
        return self.Gen.cNEBaseload_rule(m, xg, xt, self.s['Tim'],
                                         self.scenarios['Weights'],
                                         self.scenarios['WSum'],
                                         self.connections['Generation'][xh])

    def cNEFlow_rule(self, m, xt, xb, xs, xh):
        ''' Branch flows - DC model '''
        return self.ENetwork.cNEFlow_rule(m, xt, xb, xs,
                                          self.connections['Flow'][xh],
                                          self.connections['Voltage'][xh])

    def cNEFMax_rule(self, m, xt, xb, xs, xh):
        ''' Branch capacity constraint (positive) '''
        return self.ENetwork.cNEFMax_rule(m, xt, xb, xs,
                                          self.connections['Flow'][xh])

    def cNEFMin_rule(self, m, xt, xb, xs, xh):
        ''' Branch capacity constraint (positive) '''
        return self.ENetwork.cNEFMin_rule(m, xt, xb, xs,
                                          self.connections['Flow'][xh])

    def cNEGenC_rule(self, m, xg, xc, xt, xh):
        ''' Generation costs '''
        return self.Gen.cNEGenC_rule(m, xg, xc, xt,
                                     self.connections['Cost'][xh],
                                     self.connections['Generation'][xh],
                                     self.scenarios['Weights'][xt])

    def cNEGMax_rule(self, m, xg, xt, xh):
        ''' Maximum generation capacity '''
        return self.Gen.cNEGMax_rule(m, xg, xt,
                                     self.connections['Generation'][xh])

    def cNEGMin_rule(self, m, xg, xt, xh):
        ''' Minimum generation capacity '''
        return self.Gen.cNEGMin_rule(m, xg, xt,
                                     self.connections['Generation'][xh])

    def cNEGMaxUC_rule(self, m, xg, xt, xh):
        ''' Maximum generation '''
        return self.Gen.cNEGMaxUC_rule(m, xg, xt,
                                       self.connections['Generation'][xh])

    def cNEGMinDT1_rule(self, m, xg, xt, xh):
        ''' Minimum generation '''
        return self.Gen.cNEGMinDT1_rule(m, xg, xt, self.LLTime[xt],
                                        self.connections['Generation'][xh])

    def cNEGMinDT2_rule(self, m, xg, xt, xh):
        ''' Minimum generation '''
        return self.Gen.cNEGMinDT2_rule(m, xg, xt, self.LLUC,
                                        self.connections['Generation'][xh])

    def cNEGMinUC_rule(self, m, xg, xt, xh):
        ''' Minimum generation '''
        return self.Gen.cNEGMinUC_rule(m, xg, xt,
                                       self.connections['Generation'][xh])

    def cNEGMinUT1_rule(self, m, xg, xt, xh):
        ''' Minimum generation '''
        return self.Gen.cNEGMinUT1_rule(m, xg, xt, self.LLTime[xt],
                                        self.connections['Generation'][xh])

    def cNEGMinUT2_rule(self, m, xg, xt, xh):
        ''' Minimum generation '''
        return self.Gen.cNEGMinUT2_rule(m, xg, xt, self.LLUC,
                                        self.connections['Generation'][xh])

    def cNGenRampDown_rule(self, m, xg, xt, xh):
        ''' Generation ramps (down)'''
        return self.Gen.cNGenRampDown_rule(m, xg, xt, self.LLTime[xt],
                                           self.connections['Generation'][xh])

    def cNGenRampUp_rule(self, m, xg, xt, xh):
        ''' Generation ramps (up)'''
        return self.Gen.cNGenRampUp_rule(m, xg, xt, self.LLTime[xt],
                                         self.connections['Generation'][xh])

    def cNLDIni_rule(self, m, xt, xh):
        ''' Initialising dynamic loads '''
        return m.vNPump[self.connections['Pump'][xh], xt] == 0

    def cNLDMax_rule(self, m, xdl, xt, xh):
        ''' Maximum capacity of dynamic loads'''
        return (m.vNPump[self.connections['Pump'][xh]+xdl+1, xt] <=
                self.p['MaxPump'][xdl]/self.ENetwork.get_Base())

    def cNRESMax_rule(self, m, xt, xg, xh):
        # TODO: Send to pyeneD
        ''' Maximum RES generation '''
        aux = self.connections['Generation'][xh]+self.Gen.get_vNGenR(xg)
        return m.vNGen[aux, xt] <= \
            self.scenarios['RES'][self.resScenario[xg][xh]+xt] * \
            self.RES['Max'][xg]

    def cNServices_rule(self, m, xt, xh):
        ''' Provision of all services '''
        aux = self.connections['Generation'][xh]
        return sum(m.vNGen[aux+x, xt] for x in self.Gen.get_GenAll()) >= \
            sum(m.vNServ[xh*self.p['GServices']+xs, xt]
                for xs in range(self.p['GServices']))

    def cNServicesA_rule(self, m, xt, xh):
        ''' Provision of ancillary services '''
        aux = self.connections['Generation'][xh]
        return sum(m.vNGen[aux+x, xt] for x in self.Gen.get_GenAll()) >= \
            m.vNServ[xh*self.p['GServices'], xt]

    def cNServicesR_rule(self, m, xt, xh):
        ''' Provision of RES support services '''
        aux = self.connections['Generation'][xh]
        return sum(m.vNGen[aux+x, xt] for x in self.Gen.get_GenAll()) >= \
            m.vNServ[self.p['GServices']*(xh+1)-1, xt]

    def cNStore0_rule(self, m, xt):
        ''' Reference storage '''
        return m.vNStore[0, xt] == 0

    def cNStoreMax_rule(self, m, xst, xt, xh):
        ''' Maximum storage '''
        aux = xh*self.Storage['Number']+xst+1
        return m.vNStore[aux, xt] <= self.Storage['Max'][xst]

    def cNStoreMin_rule(self, m, xst, xt, xh):
        ''' Minimum storage '''
        aux = xh*self.Storage['Number']+xst+1
        return m.vNStore[aux, xt] >= 0.2*self.Storage['Max'][xst]

    def cNUncRES_rule(self, m, xt, xh):
        ''' Corrected maximum RES generation '''
        # TODO RES['Uncertainty'] is now a list
        return sum(m.vNGen[self.connections['Generation'][xh]+xg, xt]
                   for xg in self.Gen.get_GenAllR()) <= \
            sum(self.scenarios['RES'][self.resScenario[xg][xh]+xt] *
                self.RES['Max'][xg] for xg in self.s['RES']) * \
            (1-self.RES['Uncertainty'][0]) + \
            sum(m.vNFea[xh, xt] for xf in range(self.p['faux'])) + \
            m.vNServ[self.p['GServices']*(xh+1)-1, xt]

    def cNUncRES0_rule(self, m, xt, xh):
        ''' Corrected maximum RES generation without support '''
        # TODO RES['Uncertainty'] is now a list
        return sum(m.vNGen[self.connections['Generation'][xh]+xg, xt]
                   for xg in self.Gen.get_GenAllR()) <= \
            sum(self.scenarios['RES'][self.resScenario[xg][xh]+xt] *
                self.RES['Max'][xg] for xg in self.s['RES']) * \
            (1-self.RES['Uncertainty'][0]) + \
            sum(m.vNFea[xh, xt] for xf in range(self.p['faux']))

    def get_ConB(self):
        ''' Get connections between branches '''
        return self.connections['Branches']

    def get_ConC(self, x=':'):
        ''' Get connections between costs '''
        return self.connections['Cost'][x]

    def get_ConG(self, x=':'):
        ''' Get connections between generators '''
        return self.connections['Generation'][x]

    def get_ConL(self, x=':'):
        ''' Get connections between lossess constraints '''
        return self.connections['Loss'][x]

    def get_ConFea(self, x=':'):
        ''' Get connection for feasibility constraints '''
        return self.connections['Feasibility'][x]

    def get_ConP(self, x=':'):
        ''' Get connection for pump constraints '''
        return self.connections['Pump'][x]

    def get_ConS(self):
        ''' Get connections between nodes '''
        return self.connections['set']

    def get_vNGenH(self, xh, xg):
        ''' Find vNGen position of hydropower plant '''
        return self.connections['Generation'][xh]+self.Gen.get_vNGenH(xg)

    def In_From_EM(self, m, xh, xg):
        ''' Connecting  inputs from pyeneE (MWh --> MWh) '''
        aux = self.connections['Generation'][xh]+self.Gen.get_vNGenH(xg)
        return sum(m.vNGen[aux, xt]*self.scenarios['Weights'][xt]
                   for xt in self.s['Tim'])*self.ENetwork.get_Base()

    def In_From_HM(self, m, xh, xt, xg, Eff):
        ''' Connecting  inputs from pyeneH (vol --> MWh) '''
        aux = self.connections['Generation'][xh]+self.Gen.get_vNGenH(xg)
        return m.vNGen[aux, xt]*self.ENetwork.get_Base()/Eff

    def initialise(self, RM):
        ''' Initialize externally '''
        # Setting additional constraints (Security, losses and feasibilty)
        # Read network data
        self.Read()

        self.ProcessENet()

        (self.busData, self.busScenario,
         self.resScenario) = self.ProcessEDem(self.demandE)

        self.Gen.initialise(self.ENetwork, self.settings, RM)

        self.NoBuses = self.ENetwork.get_NoBus()*(1+self.NoSec2)
        self.NoBranch = self.ENetwork.get_NoBra() + \
            (self.ENetwork.get_NoBra()-1)*self.NoSec2
        self.LLStor = np.zeros((self.ENetwork.get_NoBus(),
                                self.scenarios['Number']), dtype=int)

        acu = 0
        for xh in range(self.scenarios['Number']):
            for x in range(self.Storage['Number']):
                acu += 1
                self.LLStor[self.Storage['Bus'][x]-1][xh] = acu

        self.LLTime = np.zeros(self.settings['NoTime'], dtype=int)
        self.LLTime[0] = self.settings['NoTime']-1
        for xt in range(1, self.settings['NoTime']):
            self.LLTime[xt] = xt-1

        # Initialise weights per scenario
        if self.scenarios['Weights'] is None:
            self.scenarios['Weights'] = np.ones(self.settings['NoTime'],
                                                dtype=float)

        # TODO: Redefine using pyeneD
        # Sets and parameters for modelling Ancillary service requirements
        NoSer = 0
        self.s['GAncillary'] = []
        if self.settings['Ancillary'] is not None:
            # Check for units that can provide ancillary services
            if self.conventional['Ancillary']:
                NoSer += 1
                if self.hydropower['Ancillary']:
                    ''' Conv and hydro can provide ancillary services '''
                    aux = self.Gen.get_NoCon()+self.Gen.get_NoHydro()
                    xh = self.Gen.get_NoCon()
                else:
                    # Only conventional
                    aux = self.conventional['Number']

                self.s['GAncillary'] = np.zeros(aux, dtype=int)
                for xg in range(self.Gen.get_NoCon()):
                    self.s['GAncillary'][xg] = xg
            else:
                if self.hydropower['Ancillary']:
                    # Only hydro can provide ancillary services
                    NoSer += 1
                    self.s['GAncillary'] = np.zeros(self.Gen.get_NoHydro(),
                                                    dtype=int)
                    xh = 0
                else:
                    # There are no means to provide ancillary services
                    warnings.warn('Warning: Unable to provide'
                                  ' ancillary services')

        if self.settings['Ancillary'] is not None and \
                self.hydropower['Ancillary']:
            for xg in range(self.Gen.get_NoHydro()):
                self.s['GAncillary'][xh] = self.Gen.get_NoCon()+xg
                xh += 1

        # Sets and parameters for modelling RES support
        self.s['GRES'] = []
        if self.Gen.get_NoRES() > 0 and self.RES['Uncertainty'][0] is not None:
            # Check for units that can provide RES support
            if self.conventional['RES']:
                NoSer += 1
                if self.hydropower['RES']:
                    ''' Conv and hydro can provide ancillary services '''
                    aux = self.Gen.get_NoCon()+self.Gen.get_NoHydro()
                    xh = self.Gen.get_NoCon()
                else:
                    # Only conventional
                    aux = self.Gen.get_NoCon()

                self.s['GRES'] = np.zeros(aux, dtype=int)
                for xg in range(self.Gen.get_NoCon()):
                    self.s['GRES'][xg] = xg
            else:
                if self.hydropower['RES']:
                    NoSer += 1
                    # Only hydro can provide RES services
                    self.s['GRES'] = np.zeros(self.Gen.get_NoHydro(),
                                              dtype=int)
                    xh = 0

            if self.hydropower['RES']:
                for xg in range(self.Gen.get_NoHydro()):
                    self.s['GRES'][xh] = self.Gen.get_NoCon()+xg
                    xh += 1

        # Generators providing services
        self.s['GServices'] = np.unique(np.concatenate((self.s['GAncillary'],
                                                        self.s['GRES']), 0))
        self.p['GServices'] = NoSer

    def OF_rule(self, m):
        ''' Objective function '''
        xh = self.connections['set'][0]
        return (sum((sum(m.vNGCost[self.connections['Cost'][xh]+xg, xt]
                         for xg in self.s['Gen']) +
                     sum(m.vNFea[self.connections['Feasibility'][xh]+xf, xt]
                         for xf in self.s['Fea'])*1000000) *
                    self.scenarios['Weights'][xt] for xt in self.s['Tim']) -
                sum(self.pumps['Value'][xdl]*self.ENetwork.get_Base() *
                    sum(m.vNPump[self.connections['Pump'][xh] +
                                 xdl+1, xt]*self.scenarios['Weights'][xt]
                        for xt in self.s['Tim']) for xdl in self.s['Pump']))

    def offPrint(self):
        ''' Switch off the print flags '''
        for pars in self.Print.keys():
            self.Print[pars] = False

    def print(self, m, sh=None, prnt=None, Flg=False):
        ''' Print results '''
        if sh is None:
            sh = self.s['Con']

        if prnt is None:
            from pyene.engines.pyeneP import PrintClass
            prnt = PrintClass(m, self)

        if Flg:
            print('\n\nOF: ', prnt.get_OF())
        
        for xh in sh:
            print("\n% CASE:", xh)

            if self.Print['GenBus']:
                print('\nFlow_EGen_Bus=', self.Gen.get_GenDataAll(), ';')

            if self.Print['Generation']:
                print("\nFlow_EGen=[")
                for xn in range(self.Gen.get_NoGen()):
                    x1 = self.connections['Generation'][xh]+xn
                    for x2 in range(self.settings['NoTime']):
                        print("%8.4f " % prnt.get_GenerationP(x1, x2), end='')
                    print()
                print("];")
                if prnt.type == 2:
                    print("\nFlow_EGenQ=[")
                    for xn in range(self.Gen.get_NoGen()):
                        x1 = self.connections['Generation'][xh]+xn
                        for x2 in range(self.settings['NoTime']):
                            print("%8.4f " % prnt.get_GenerationQ(x1, x2), end='')
                        print()
                    print("];")
                    

            if self.Print['UC']:
                print("\nBin_EGen=[")
                for xn in range(self.Gen.get_NoGen()):
                    if self.Gen.get_Bin(xn) is None:
                        x1 = self.connections['Generation'][xh]
                    else:
                        x1 = self.Gen.get_Bin(xn) + \
                            self.connections['Generation'][xh]
                    for x2 in range(self.settings['NoTime']):
                        print("%2.0f " % prnt.get_GenerationUC(x1, x2), end='')
                    print()
                print("];")

            if self.Print['Flows'] and self.settings['Flag']:
                print("\nFlow_EPower=[")
                for xb in range(self.ENetwork.get_NoBra()):
                    x1 = self.connections['Flow'][xh]+xb
                    for x2 in range(self.settings['NoTime']):
                        print("%8.4f " % prnt.get_Line_FlowP0(x1, x2), end='')
                    print()
                print("];")
                if prnt.type == 2:
                    print("\nFlow_EPowerP1=[")
                    for xb in range(self.ENetwork.get_NoBra()):
                        x1 = self.connections['Flow'][xh]+xb
                        for x2 in range(self.settings['NoTime']):
                            print("%8.4f " % prnt.get_Line_FlowP1(x1, x2), end='')
                        print()
                    print("];")
                    print("\nFlow_EPowerQ0=[")
                    for xb in range(self.ENetwork.get_NoBra()):
                        x1 = self.connections['Flow'][xh]+xb
                        for x2 in range(self.settings['NoTime']):
                            print("%8.4f " % prnt.get_Line_FlowQ0(x1, x2), end='')
                        print()
                    print("];")
                    print("\nFlow_EPowerQ1=[")
                    for xb in range(self.ENetwork.get_NoBra()):
                        x1 = self.connections['Flow'][xh]+xb
                        for x2 in range(self.settings['NoTime']):
                            print("%8.4f " % prnt.get_Line_FlowQ1(x1, x2), end='')
                        print()
                    print("];")

            if self.Print['Voltages'] and self.settings['Flag']:
                print("\nVoltage_Angle=[")
                for xn in range(self.NoBuses):
                    x1 = self.connections['Voltage'][xh]+xn
                    for x2 in range(self.settings['NoTime']):
                        print("%8.4f " % prnt.get_Voltage_ang(x1, x2), end='')
                    print()
                print("];")
                if prnt.type == 2:
                    print("\nVoltage_Magnitude=[")
                    for xn in range(self.NoBuses):
                        x1 = self.connections['Voltage'][xh]+xn
                        for x2 in range(self.settings['NoTime']):
                            print("%8.4f " % prnt.get_Voltage_pu(x1, x2), end='')
                        print()
                    print("];")

            if self.Print['Losses'] and self.settings['Flag']:
                print("\nEPower_Loss=[")
                (LossDt, LossDtQ) = self.printLosses(prnt, xh)
                for xb in range(self.ENetwork.get_NoBra()):
                    for xt in range(self.settings['NoTime']):
                        print("%8.4f " % LossDt[xb][xt], end='')
                    print()
                print("];")
                if prnt.type == 2:
                    print("\nEPower_LossQ=[")
                    for xb in range(self.ENetwork.get_NoBra()):
                        for xt in range(self.settings['NoTime']):
                            print("%8.4f " % LossDtQ[xb][xt], end='')
                        print()
                    print("];")

            if self.Print['Pump']:
                print("\nPumps=[")
                for xp in range(self.pumps['Number']):
                    x1 = self.connections['Pump'][xh]+xp
                    for x2 in range(self.settings['NoTime']):
                        print("%8.4f " % prnt.get_Pumps(x1, x2), end='')
                    print()
                print("];")

            if self.Print['Curtailment']:
                print("\nFeas=[")
                for xn in range(self.ENetwork.get_NoBus()):
                    if self.p['LLFea1'][xn] == 0:
                        for x2 in range(self.settings['NoTime']):
                            print("%8.4f " % 0, end='')
                    else:
                        x1 = self.connections['Feasibility'][xh] + \
                            self.p['LLFea2'][xn]
                        for x2 in range(self.settings['NoTime']):                                                                        
                            print("%8.4f " % prnt.get_Curtailment(x1, x2),
                                  end='')
                    print()
                print("];")

            if self.Print['Services'] and len(self.s['GServices']) > 0:
                print("\nServ=[")
                for xs in range(self.p['GServices']):
                    x1 = self.p['GServices']*xh+xs
                    for xt in range(self.settings['NoTime']):
                        print("%8.4f " % prnt.get_Services(x1, x2), end='')
                    print()
                print("];")

    def printLosses(self, prnt, xh):
        ''' Get losses data to be printed '''
        LossesDt = np.zeros((self.ENetwork.get_NoBra(),
                             self.settings['NoTime']), dtype=float)
        LossesDtQ = np.zeros((self.ENetwork.get_NoBra(),
                             self.settings['NoTime']), dtype=float)
        
        if self.settings['Loss'] is None:
            aux = 0
            for xb in range(self.ENetwork.get_NoBra()):
                x1 = self.connections['Loss'][xh]+xb
                for x2 in range(self.settings['NoTime']):
                    LossesDt[xb][x2] = prnt.get_Line_LossP(x1, x2)
                    if prnt.type == 2:
                        LossesDtQ[xb][x2] = prnt.get_Line_LossQ(x1, x2)
        else:
            # Get all losses
            FullLoss = np.zeros(self.settings['NoTime'], dtype=float)
            for xt in self.s['Tim']:
                # Get all power generation
                for xn in range(self.Gen.get_NoGen()):
                    x1 = self.connections['Generation'][xh]+xn
                    FullLoss[xt] += prnt.get_GenerationP(x1, xt)

                # Substract demand
                for xn in self.s['Bus']:
                    FullLoss[xt] -= self.busData[xn]*self.scenarios['Demand'] \
                        [xt*self.p['daux']+self.busScenario[xn][xh]] * \
                        self.ENetwork.get_Base()

                    # Adding curtailment (dummy generators)
                    x1 = self.connections['Feasibility'][xh] + \
                        self.p['LLFea2'][xn]
                    FullLoss[xt] += prnt.get_Curtailment(x1, xt)

                # Substract non-technical losses
                for xb in range(self.ENetwork.get_NoBra()):
                    FullLoss[xt] -= self.ENetwork.Branch[xb].getLoss()

            # Allocate losses per line
            for xt in self.s['Tim']:
                FullFlow = 0
                for xb in range(self.ENetwork.get_NoBra()):
                    FullFlow += prnt.get_Line_FlowP0(self.connections['Flow'][xh]+xb, xt)

                if FullFlow > 0:
                    for xb in range(self.ENetwork.get_NoBra()):
                        aux = abs(prnt.get_Line_FlowP0(self.connections['Flow']
                                                       [xh]+xb, xt))/FullFlow
                        LossesDt[xb][xt] = FullLoss[xt] * aux + \
                            self.ENetwork.Branch[xb].getLoss()

        return LossesDt, LossesDtQ
        
    def ProcessEDem(self, ENetDem):
        ''' Process demand and generation parameters '''
        # Adjust demand profiles
        busData = np.zeros(self.ENetwork.get_NoBus(), dtype=float)
        for xn in range(self.ENetwork.get_NoBus()):
            busData[xn] = (self.ENetwork.Bus[xn].getLoss() +
                           self.demandE['PD'][xn])/self.ENetwork.get_Base()

        # Auxiliar to find demand profiles
        busScenario = np.zeros((self.ENetwork.get_NoBus(),
                                self.scenarios['Number']), dtype=int)

        if self.scenarios['NoDem'] > 0:
            acu = 0
            for xs in range(self.scenarios['Number']):
                for xn in range(self.ENetwork.get_NoBus()):
                    busScenario[xn][xs] = ((self.scenarios['Links'][acu]-1) *
                                           self.settings['NoTime'])
                    acu += 1

        # Auxiliar to find RES profiles
        resScenario = np.zeros((self.Gen.get_NoRES(),
                                self.scenarios['Number']), dtype=int)
        for xh in range(self.scenarios['Number']):
            for xg in range(self.Gen.get_NoRES()):
                # Profile location
                resScenario[xg][xh] = self.settings['NoTime'] * \
                    (self.scenarios['LinksRes'][xg+xh*self.Gen.get_NoRES()]-1)

        return (busData, busScenario, resScenario)

    def ProcessENet(self):
        ''' Process information for optimisation purposes '''

        # Initialise electricity network object
        self.ENetwork.initialise(self.settings)

        # Add security constraints
        if len(self.settings['Security']) == 0 and \
                self.settings['SecurityFlag']:
            self.settings['Security'] = \
                [x+1 for x in range(self.connections['Branches'])]

        # Add security considerations
        NoSec2 = len(self.settings['Security'])

        # Number of parameters required for simulating security
        aux = self.connections['Branches']
        NoSec1 = self.connections['Branches']*(1+NoSec2)-NoSec2
        # Auxiliaries for modelling security considerations
        # Position of the variables
        LLESec1 = np.zeros((NoSec1, 2), dtype=int)

        for xb in range(self.connections['Branches']):
            LLESec1[xb][0] = xb
        aux = self.connections['Branches']
        x0 = aux
        xacu = 0
        for xs in range(NoSec2):
            xacu += self.ENetwork.get_NoBus()
            for xb in range(self.connections['Branches']):
                if xb+1 != self.settings['Security'][xs]:
                    LLESec1[x0][:] = [xb, xacu]
                    x0 += 1
        self.NoSec1 = NoSec1
        self.NoSec2 = NoSec2
        self.p['LLESec1'] = LLESec1

        # Add piece-wise power losses estimation
        if self.settings['Losses']:
            # Auxiliar for the cuadratic function to model losses
            # Choosing points for the lines
            aux = self.connections['Loss_Param']

            # Number of points to model
            Number_LossCon = len(aux)-1

            # The function is modelled between the points a<b) as:
            # a^2-a*(b^2-a^2)/(b-a) + x *(b^2-a^2)/(b-a)
            Loss_Con1 = np.zeros(Number_LossCon, dtype=float)
            Loss_Con2 = np.zeros(Number_LossCon, dtype=float)
            for x1 in range(Number_LossCon):
                Loss_Con2[x1] = ((aux[x1+1]**2-aux[x1]**2) /
                                 (aux[x1+1]-aux[x1]))
                Loss_Con1[x1] = aux[x1]**2-aux[x1]*Loss_Con2[x1]

            self.Number_LossCon = Number_LossCon
            self.ENetwork.loss['A'] = Loss_Con1
            self.ENetwork.loss['B'] = Loss_Con2

        # Add LL for dynamic loads
        LLDL = np.zeros(self.ENetwork.get_NoBus(), dtype=int)
        for xdl in range(self.pumps['Number']):
            LLDL[self.pumps['Bus'][xdl]-1] = xdl+1
        self.p['LLPump'] = LLDL

        # Add LL for feasibility constraints only in nodes with demand
        NoFea = 0
        LLFea1 = np.zeros(self.ENetwork.get_NoBus(), dtype=int)
        LLFea2 = np.zeros(self.ENetwork.get_NoBus(), dtype=int)
        for xn in range(self.ENetwork.get_NoBus()):
            if self.demandE['PD'][xn] > 0:
                LLFea1[xn] = 1
                LLFea2[xn] = NoFea
                NoFea += 1
        self.p['LLFea1'] = LLFea1
        self.p['LLFea2'] = LLFea2
        self.NoFea = NoFea

    def Read(self):
        ''' Read input data '''
        # Load file
        mpc = json.load(open(self.settings['File']))

        GenNCost = np.array(mpc['gencost']['COST'], dtype=int)
        NoOGen = len(GenNCost)

        # Defining device classes
        from pyensys.engines.pyeneD import ElectricityNetwork, Generators

        # Define network model
        # The definition and configuration methods are separated so, in
        # principle the classes can be manually configured
        self.ENetwork = ElectricityNetwork(mpc['NoBus'], mpc["NoBranch"])
        self.ENetwork.MPCconfigure(mpc)

        # Define generator model
        self.Gen = Generators(NoOGen, self.hydropower['Number'],
                              self.RES['Number'])

        self.Gen.MPCconfigure(mpc, self.conventional, self.hydropower,
                              self.RES)
        # TODO: Remove conventional, hydropower and RES
        # del self.conventional, self.hydropower, self.RES

        self.connections['Branches'] = mpc['NoBranch']
        self.demandE = {
                'PD': np.array(mpc['bus']['PD'], dtype=float),
                'QD': np.array(mpc['bus']['QD'], dtype=float),
                }

        # Adjust demand dimensions
        self.scenarios['Demand'] = np.asarray(self.scenarios['Demand'])

        # Default settings for demand profiles
        if self.scenarios['Links'] == 'Default':
            self.scenarios['Links'] = np.ones(self.ENetwork.get_NoBus() *
                                              self.scenarios['Number'],
                                              dtype=int)
            acu = self.ENetwork.get_NoBus()
            for xs in range(self.scenarios['Number']-1):
                for xt in range(self.ENetwork.get_NoBus()):
                    self.scenarios['Links'][acu] = xs+2
                    acu += 1

        # Default settings for RES profiles
        # All devices are linked to the same profile
        if self.scenarios['NoRES'] == 1:
            self.scenarios['LinksRes'] = np.ones(self.scenarios['Number'] *
                                                 self.Gen.get_NoRES(),
                                                 dtype=int)
        # i.e., each scenario is linked to a profile
        elif self.scenarios['LinksRes'] == 'Default':
            self.scenarios['LinksRes'] = np.ones(self.scenarios['Number'] *
                                                 self.Gen.get_NoRES(),
                                                 dtype=int)
            acu = self.Gen.get_NoRES()
            for xs in range(self.scenarios['Number']-1):
                for xt in range(self.Gen.get_NoRES()):
                    self.scenarios['LinksRes'][acu] = xs+2
                    acu += 1

    def set_ConS(self, val):
        ''' Set connections for nodes '''
        self.connections['set'] = val

    def set_ConF(self, val):
        ''' Set connections between branch flows '''
        self.connections['Flow'] = val

    def set_ConV(self, val):
        ''' Set connections between voltages '''
        self.connections['Voltage'] = val

    def set_ConL(self, val):
        ''' Set connections between branch Losses '''
        self.connections['Loss'] = val

    def set_ConG(self, val):
        ''' Set connections between generators '''
        self.connections['Generation'] = val

    def set_ConC(self, val):
        ''' Set connections between costs '''
        self.connections['Cost'] = val

    def set_ConP(self, val):
        ''' Set connections between pumps '''
        self.connections['Pump'] = val

    def set_ConFea(self, val):
        ''' Set connections between feasibility constraints '''
        self.connections['Feasibility'] = val



""" Test the pyeneE engine. """
from pyene.fixtures import testConfig, json_directory
from pyene.engines.pyene import pyeneClass as pe
import os


def test_pyeneH_Time():
    ''' Check time dependence in default network '''
    print('test_pyeneH_Time')
    # Load default configuration and add 4 nodes network
    conf = testConfig()
    conf.HM.settings['File'] = os.path.join(json_directory(), 'caseH4.json')
    EN = pe(conf.EN)
    conf.HM.settings['Flag'] = True
    (HM, HModel, results) = EN.HSim(conf)
    HM.print(HModel)

    assert 0.0001 >= abs(HModel.OF.expr()-6766.6666) and \
        0.0001 >= abs(HModel.vHout[0, 0].value-66.6680) and \
        0.0001 >= abs(HModel.vHout[0, 1].value-305.5747) and \
        0.0001 >= abs(HModel.vHout[0, 2].value-171.7524) and \
        0.0001 >= abs(HModel.vHout[1, 2].value-66.6668) and \
        0.0001 >= abs(HModel.vHout[1, 3].value-236.2205) and \
        0.0001 >= abs(HModel.vHout[1, 4].value-103.9565) and \
        0.0001 >= abs(HModel.vHout[2, 2].value-100.0001) and \
        0.0001 >= abs(HModel.vHout[2, 3].value-354.3308) and \
        0.0001 >= abs(HModel.vHout[2, 4].value-155.9348)


def test_pyeneH_Scenarios():
    ''' Check multiple scenario analysis '''
    print('test_pyeneH_Scenarios')
    conf = testConfig()
    conf.HM.settings['File'] = os.path.join(json_directory(), 'caseH4.json')
    conf.HM.settings['Flag'] = True
    conf.HM.settings['NoTime'] = 5  # Five periods
    conf.HM.connections['Number'] = 2  # Two scenarios
    # Ading additional water in both scenarios
    conf.HM.settings['In'] = [[0, 1, 600], [3, 3, 800]]
    # Connecting the reservoirs at the beginning and end of each scenario
    conf.HM.connections['LinksF'] = [[0, 0], [1, 0]]
    conf.HM.connections['LinksT'] = [[0, 1], [1, 1]]
    EN = pe(conf.EN)
    (HM, HModel, results) = EN.HSim(conf)
    HM.print(HModel)
    print('OF: ', HModel.OF.expr())

    assert 0.0001 >= abs(HModel.OF.expr()-3500) and \
        0.0001 >= abs(HModel.vHout[0, 1].value-325.0060) and \
        0.0001 >= abs(HModel.vHout[0, 2].value-183.6914) and \
        0.0001 >= abs(HModel.vHout[4, 3].value-239.9525) and \
        0.0001 >= abs(HModel.vHout[4, 4].value-106.0274)


def test_pyeneH_ScenParts():
    ''' Check multiple scenario analysis '''
    print('test_pyeneH_ScenParts')
    conf = testConfig()
    conf.HM.settings['File'] = os.path.join(json_directory(), 'caseH4.json')
    conf.HM.settings['Flag'] = True
    conf.HM.settings['NoTime'] = 5  # Five periods
    conf.HM.connections['Number'] = 2  # Two scenarios
    # Ading additional water in both scenarios
    conf.HM.settings['In'] = [[0, 1, 600], [3, 3, 800]]
    # Connecting the reservoirs at the beginning and end of each scenario
    conf.HM.connections['LinksF'] = [[0, 0], [1, 0]]
    conf.HM.connections['LinksT'] = [[0, 1], [1, 1]]
    conf.HM.rivers['Parts'] = [5]  # Model each river using several parts
    print()
    print(conf.HM.rivers)
    print()
    EN = pe(conf.EN)
    (HM, HModel, results) = EN.HSim(conf)
    HM.print(HModel)
    print('OF: ', HModel.OF.expr())

    assert 0.0001 >= abs(HModel.OF.expr()-3500) and \
        0.0001 >= abs(HModel.vHout[0, 1].value-383.7028) and \
        0.0001 >= abs(HModel.vHout[0, 2].value-124.6046) and \
        0.0001 >= abs(HModel.vHout[4, 3].value-259.6517) and \
        0.0001 >= abs(HModel.vHout[4, 4].value-83.0653)



"""
Created on Mon April 06 2020

This python file containts the classes and methods for the analysis and
modifications of the topology and electrical characteristics of power system.
Furthermore, tools to build the temporal tree are provided

@author: Dr. Jose Nicolas Melchor Gutierrez
"""

import networkx as nx
import logging
import copy
from .pyene_Parameters import ElectricityNetwork, Bus, TransmissionLine

class PowerSystemIslandsIsolations(ElectricityNetwork):
    ''' This class contains all necessary methods to find all islands and 
    isolated nodes in a network '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        __data2 = {
            'IsolatedNodes': [], # Isolated nodes in the power system
            'IsolatedGenerators': {}, # Isolated generators in the system
            'Islands': [], # Islands in the whole power system
            'Disconnectedserieselements' : {} # This array contains all 
                # disconnected series elements (e.g. transmission lines and 
                # transformers)
        }
        self._data.update(__data2)
        del __data2
        logging.basicConfig(format='%(asctime)s %(message)s', \
            datefmt='%m/%d/%Y %I:%M:%S %p', level=logging.INFO)
    
    def islands_isolations(self):
        ''' This class method calls and controls all main methods in this class'''
        auxp = 'Running network analyser - Determining islands and \
            isolated nodes in the power system'
        logging.info(" ".join(auxp.split()))
        copy_electricity_network = copy.deepcopy(self)
        self.__delete_disconnected_series_elements()
        self.find_isolates()
        self.find_islands()
        self.set_electricity_network_data(ob=copy_electricity_network)

    def edges_graph(self, graph=None):
        ''' This class method load all edges in the graph'''
        auxp = "No edge data has been loaded in the \
                class PowerSystemIslandsIsolations"
        are_edges = False
        for xseries in self.get_series_elements_names():
            aux = self.get_object_elements(name_object=xseries, \
                name_element='position',  pos_object=0)
            if aux != None:
                are_edges = True
                break
        assert are_edges, " ".join(auxp.split())
        assert graph.is_multigraph(), "graph is not a multigraph"
        # Edges for transmission lines
        for xseries in self.get_series_elements_names():
            aux = self.get_objects(name=xseries)
            for edges in aux:
                aux1 = edges.get_element(name='bus_position')
                graph.add_edge(aux1[0], aux1[1])
        return graph
    
    def nodes_graph(self, graph=None):
        ''' This class method load all nodes in the graph'''
        auxp = "No bus data has been loaded in the class \
            PowerSystemIslandsIsolations"
        aux = self.get_object_elements(name_object='bus', \
            name_element='position',  pos_object=0)
        assert aux != None, " ".join(auxp.split())
        assert graph.is_multigraph(), "graph is not a multigraph"

        aux = self.get_objects(name='bus')
        for nodes in aux:
            aux1 = nodes.get_element(name='position')
            graph.add_node(aux1, obj=nodes)
        return graph
    
    def find_islands(self):
        ''' This class method finds all islands in the whole power system '''
        G = nx.MultiGraph()
        G = self.nodes_graph(G)
        G = self.edges_graph(G)
        if nx.number_connected_components(G) > 0:
            S = [G.subgraph(c).copy() for c in \
                nx.connected_components(G)]
            self._data['Islands'] = [ElectricityNetwork() for _ in \
                range(len(S))]
            self.__add_nodes_to_island(S)
            for xgentypes in self.get_generation_types_names():
                self.__add_object_to_islands(name_object=xgentypes)
            for xseriestype in self.get_series_elements_names():
                self.__add_object_to_islands(name_object=xseriestype)
            self.__add_parameters_to_islands()
            self.__update_all_pos_islands()
            auxp = 'Network analyser message - the power system under \
                analysis has {0} islands'.format(\
                str(self.get_no_elements(name='Islands')))
            logging.info(" ".join(auxp.split()))
        else:
            auxp = 'Network analyser message - the power system under \
                analysis does not have islands'
            logging.info(" ".join(auxp.split()))

    def find_isolates(self):
        ''' This class method finds all isolated nodes in the power system '''
        G = nx.MultiGraph()
        G = self.nodes_graph(G)
        G = self.edges_graph(G)
        isolatednodesgraph = list(nx.isolates(G))
        if isolatednodesgraph != []:
            G.remove_nodes_from(isolatednodesgraph)
            isolatednodesgraph.sort()
            self._data['IsolatedNodes'], self._data['IsolatedGenerators'] = \
                self.__delete_nodes_graph(isolatednodesgraph)
            auxp = 'Network analyser message - the power system under \
                analysis has {0} isolated nodes'.format(\
                len(self._data['IsolatedNodes']))
            logging.info(" ".join(auxp.split()))
        else:
            auxp = 'Network analyser message - the power system under \
                analysis does not have isolated nodes'
            logging.info(" ".join(auxp.split()))

    def __add_nodes_to_island(self, graphs=None, lt=None):
        ''' This class method add the nodes to the islands '''
        assert isinstance(self._data['Islands'], list) and \
            isinstance(self._data['Islands'][0], ElectricityNetwork), \
                "List of islands does not have the necessary format"
        if isinstance(graphs, list):
            assert graphs[0].is_multigraph(), "graphs is not a multigraph"
            for aux in range(self.get_no_elements(name='Islands')):
                aux1 = [aux2[1]['obj'] for aux2 in \
                    graphs[aux].nodes(data=True)]
                self._data['Islands'][aux].set_objects(name='bus', \
                    list_obj=aux1)
                del aux1

    def __add_object_to_islands(self, name_object=None):
        ''' This class method add the transmission lines to the islands '''
        auxp = "List of islands does not have the necessary format"
        assert isinstance(self._data['Islands'], list) and \
            isinstance(self._data['Islands'][0], ElectricityNetwork), \
            " ".join(auxp.split())
        auxp = "No valid objects name has been passed to the function \
            __add_object_to_islands \
            in the class {0}".format(self.__class__.__name__)
        assert name_object is not None, " ".join(auxp.split())
        for aux1 in range(self.get_no_elements(name='Islands')):
            aux_objects = []
            for aux2 in self._data['Islands'][aux1].get_objects(name='bus'):
                if isinstance(aux2.get_element(\
                    name=name_object+'_position'), list):
                    aux_objects.extend(aux2.get_element(\
                        name=name_object+'_position'))
                else:
                    aux_objects.append(aux2.get_element(\
                        name=name_object+'_position'))
            aux_objects = list(dict.fromkeys(aux_objects))
            if self.get_objects(name=name_object, pos=aux_objects) != []:
                self._data['Islands'][aux1].set_objects(name=name_object, \
                    list_obj=self.get_objects(name=name_object, pos=aux_objects))
    
    def __add_parameters_to_islands(self):
        ''' This class method set the parameters of the electricity network on 
        each island '''
        for aux1 in range(self.get_no_elements(name='Islands')):
            for xparam in self.get_parameters_list():
                self._data['Islands'][aux1].set_element(name=xparam, val=\
                    self.get_element(name=xparam))

    def __delete_nodes_graph(self, nodes=None):
        ''' This class method remove the isolated nodes from the graph. The 
        method return a list with all information of the deleted nodes'''
        assert isinstance(nodes, list), "Isolated nodes are not an empty list"
        aux = range(len(nodes))
        isolated_nodes = [Bus() for _ in aux]
        isolated_generators = {}
        for auxgen in self.get_generation_types_names():
            isolated_generators[auxgen] = []
        for aux1 in aux:
            isolated_nodes[aux1] = self.get_objects(name='bus', pos=nodes[aux1])
            for auxgen in self.get_generation_types_names():
                gen_info = self.get_object_elements(name_object='bus', \
                    name_element=auxgen+'_position',pos_object=nodes[aux1])
                if gen_info != []:
                    for gen_iter in gen_info:
                        isolated_generators[auxgen].append(self.get_objects(\
                            name=auxgen,pos=gen_iter))

        self.delete_objects(name='bus', pos=nodes)
        return isolated_nodes, isolated_generators

    def __delete_disconnected_series_elements(self):
        ''' This class method deletes all disconnected transmission lines, 
        transformers and all series elements from the network '''
        for xseries in self.get_series_elements_names():
            auxremovals = []
            for xelement in self.get_objects(name=xseries):
                if xelement.get_element('status') == 0:
                    auxremovals.append(xelement.get_element('position'))
            if auxremovals != []:
                auxstore = {xseries : self.get_objects(name=xseries, pos=\
                    auxremovals)}
                self._data['Disconnectedserieselements'].update(auxstore)
                self.delete_objects(name=xseries, pos=auxremovals)

    def __update_all_pos_islands(self):
        ''' Update the position of all nodes, transmission lines, etc. on 
        each island '''
        for xisl in range(self.get_no_elements(name='Islands')):
            self._data['Islands'][xisl].update_all_positions()

class PowerSystemReduction(ElectricityNetwork):
    ''' This class contains all necessary methods to reduce a network depending 
    on the requirements of the user
    
    The options considered are:

    1. Simplify power system network until a desired voltage level - network 
        characteristics of the reduced network are omitted
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        __data2 = {
            'VoltageLevels': [], # Voltage levels in the power system
            'Supernodeinitial' : None,      # Initial node for supernodes
            'Supernodelineinitial' : None,  # Initial artificial line for 
                                            # supernodes
            'SupernodesNetworkInfo': [],     # Network info of supernodes
            'Interconnection_branches': [], # Original branches that connects to supernodes
                # This is containts a tuple with the following order:
                # Position 0: Position in original list of branches
                # Position 1: From bus original
                # Position 2: To bus original
                # Position 3: Position in new list of branches
                # Position 4: From bus new
                # Position 5: To bus new
            'Preserved_branches': [], # Preserved branches in the system with original numeration
                # This is containts a tuple with the following order:
                # Position 0: Position in original list of branches
                # Position 1: From bus original
                # Position 2: To bus original
                # Position 3: Position in new list of branches
                # Position 4: From bus new
                # Position 5: To bus new
            'Removed_branches': [], # Original branches that connects to supernodes
                # This is containts a tuple with the following order:
                # Position 0: Position in original list of branches
                # Position 1: From bus original
                # Position 2: To bus original
        }
        self._data.update(__data2)
        del __data2
        logging.basicConfig(format='%(asctime)s %(message)s', \
            datefmt='%m/%d/%Y %I:%M:%S %p', level=logging.INFO)
    
    def network_reduction_voltage_no_characteristics(self, vol_kv=None, \
        initial_bus_supernode=None, initial_artificial_line=None):
        ''' This class method controls the functions to reduce the network \
        until the desired voltage level (vol_kv) without including 
        electrical characteristics of the eliminated elements

        The vol parameter needs to be in kV
        The electricity_network parameter needs to be an object of the 
            ElectricityNetwork class
        '''
        assert vol_kv is not None, "No voltage passed to reduce the network"
        self.__find_voltage_levels()
        auxp = "The indicated voltage level is not in the list of voltage \
            levels. The valid voltages are: {0}".format(\
            self._data['VoltageLevels'])
        assert vol_kv in self._data['VoltageLevels'], " ".join(auxp.split())
        flags = {
            'bus' : [False for _ in \
            range(self.get_no_objects(name='bus'))], # Boolean list of Bus 
                # objects
            'transmissionline' : [False for _ in \
            range(self.get_no_objects(name='transmissionline'))], # Boolean 
                # list of transmission line objects
            'transformers' : [False for _ in \
            range(self.get_no_objects(name='transformers'))], # Boolean 
                # list of two winding trafo objects
            }
        flag_continue=True
        counter = 0
        supernodes = []
        while flag_continue:
            flag_continue = False
            for xaux in range(self.get_no_objects(name='bus')):
                if not flags['bus'][xaux]:
                    supernode = { # This dictionary contains the positions of the elements
                        # of the network that will be converted into a supernode
                        'bus' : [], # list of Bus objects in the supernode
                        'coupling_buses' : [], # list with the positions of nodes in the 
                            # supernode that interconnect the supernode with the reduced
                            # network
                        'coupling_bus_name' : None, # name of the coupling node  
                            # that interconnect the supernode with the reduced network
                        'transmissionline' : [], # list of transmission line objects in the
                            # supernode
                        'transformers' : [], # list of trafo objects in the
                            # supernode
                        'equivalent_active_power_demand' : None, # Equivalent active power demand
                            # to be connected to the supernode
                        'equivalent_reactive_power_demand' : None, # Equivalent reactive power demand
                            # to be connected to the supernode
                        'artificial_lines_info' : [], # transformers are replaced with artificial 
                            # lines with the same electrical characteristics
                        'equivalent_non_technical_losses_fix' : None,
                        'zone_supernode' : None
                    }
                    flags, supernode, flag_supernode = self.__voltage_track(\
                        bus_position=xaux, vol_kv=vol_kv, \
                        flags=flags, supernode=supernode, flag_supernode=False)
                    if flag_supernode:
                        supernode['coupling_bus_name'] = \
                            'supernode_'+str(counter)
                        counter += 1
                        supernodes.append(supernode)
                    flag_continue = True
        if supernodes != []:
            auxp = 'Network analyser message - the power system under \
                analysis has {0} supernodes'.format(len(supernodes))
            logging.info(" ".join(auxp.split()))
            auxp = 'Network analyser message - the power system under \
                analysis has {0} nodes before reduction'.format(\
                self.get_no_objects('bus'))
            logging.info(" ".join(auxp.split()))
            auxp = 'Network analyser message - the power system under \
                analysis has {0} transmission lines before reduction'.format(\
                self.get_no_objects('transmissionline'))
            logging.info(" ".join(auxp.split()))
            auxp = 'Network analyser message - the power system under \
                analysis has {0} transformers before reduction'.format(\
                self.get_no_objects('transformers'))
            logging.info(" ".join(auxp.split()))
            self.__reduce_network_from_supernodes(supernodes=supernodes)
            self.__add_artificial_elements(supernodes=supernodes, \
                initial_bus_supernode=initial_bus_supernode, \
                initial_artificial_line=initial_artificial_line)
            auxp = 'Network analyser message - the power system under \
                analysis has {0} nodes after reduction'.format(\
                self.get_no_objects('bus'))
            logging.info(" ".join(auxp.split()))
            auxp = 'Network analyser message - the power system under \
                analysis has {0} transmission lines after reduction'.format(\
                self.get_no_objects('transmissionline'))
            logging.info(" ".join(auxp.split()))
            auxp = 'Network analyser message - the power system under \
                analysis has {0} transformers after reduction'.format(\
                self.get_no_objects('transformers'))
            logging.info(" ".join(auxp.split()))
        else:
            auxp = 'Network analyser message - the power system under \
                analysis has not been reduced - check if the indicated voltage \
                    {0} is correct'.format(vol_kv)
            logging.info(" ".join(auxp.split()))
        return supernodes
                    
    def reduction(self):
        ''' This is the main class method'''
        self.network_reduction_voltage_no_characteristics()
    
    def __add_artificial_elements(self, supernodes=None, \
        initial_bus_supernode=None, initial_artificial_line=None):
        ''' This class method add the artificial buses and lines to the 
        network '''
        assert isinstance(supernodes, list), "Incorrect \
            object passed for supernodes"
        copy_nodes = self.get_objects(name='bus')
        if initial_bus_supernode == None:
            number_nodes = []
            for xaux in copy_nodes:
                number_nodes.append(xaux.get_element(name='number'))
            for xaux in self._data['SupernodesNetworkInfo']:
                for xaux1 in xaux['bus']:
                    number_nodes.append(xaux1.get_element(name='number'))
            aux = 1
            while aux <= max(number_nodes):
                aux *= 10
        else:
            aux = initial_bus_supernode
        
        copy_lines = self.get_objects(name='transmissionline')
        if initial_artificial_line == None:
            number_lines = []
            for xaux in copy_lines:
                number_lines.append(xaux.get_element(name='number'))
            for xaux in self._data['SupernodesNetworkInfo']:
                for xaux1 in xaux['transmissionline']:
                    number_lines.append(xaux1.get_element(name='number'))
            auxline = 1
            while auxline <= max(number_lines):
                auxline *= 10
        else:
            auxline = initial_artificial_line

        new_nodes = []
        new_lines = []
        counternodes = 0
        counterlines = 0
        for x in range(len(supernodes)):
            if len(supernodes[x]['coupling_buses']) > 1:
                if counternodes == 0:
                    auxp = 'Network analyser message - artificial nodes with \
                        numbering starting from {0} are added to the \
                        system'.format(aux)
                    logging.info(" ".join(auxp.split()))
                    auxp = 'Network analyser message - artificial lines with \
                        numbering starting from {0} are added to the \
                        system'.format(auxline)
                    if self._data['Supernodeinitial'] == None:
                        self._data['Supernodeinitial'] = aux
                    if self._data['Supernodelineinitial'] == None:
                        self._data['Supernodelineinitial'] = auxline
                    logging.info(" ".join(auxp.split()))
                new_nodes.append(Bus())
                new_nodes = self.__add_artificial_node(supernodes=supernodes, \
                    counter=counternodes, number_node=aux, new_nodes=new_nodes,\
                    pos_supernode=x)
                self.__add_generators_to_artificial_node(supernodes=supernodes,\
                    counter=counternodes, number_node=aux, new_nodes=new_nodes,\
                    pos_supernode=x)
                counterlines, auxline = self.__add_artificial_lines(\
                    supernodes=supernodes, counter=counterlines, \
                    number_line=auxline, new_nodes=new_nodes, \
                    counternode=counternodes, number_node=aux, pos_supernode=x,
                    new_lines=new_lines)
                counternodes += 1
                aux += 1
            else:
                for xnode in copy_nodes:
                    if xnode.get_element(name='number') == \
                        self._data['SupernodesNetworkInfo'][x]\
                        ['coupling_buses'][0].get_element(name='number'):
                        xnode.set_element(name='active_power_demand_peak',\
                            val=supernodes[x]\
                                ['equivalent_active_power_demand'] + \
                            xnode.get_element(name='active_power_demand_peak'))
                        xnode.set_element(name='reactive_power_demand_peak',\
                            val=supernodes[x]\
                                ['equivalent_reactive_power_demand'] + \
                            xnode.get_element(\
                                name='reactive_power_demand_peak'))
                        xnode.set_element(name='non_technical_losses_fix',\
                            val=supernodes[x][\
                                'equivalent_non_technical_losses_fix']+
                            xnode.get_element(\
                                name='non_technical_losses_fix'))
                        self.__add_generators_to_artificial_node(\
                            supernodes=supernodes, counter=\
                            xnode.get_element(name='position'), \
                            number_node=xnode.get_element(name='number'), \
                            new_nodes=copy_nodes, \
                            pos_supernode=x)
        
        aux_num_lines = len(copy_lines)
        for line in self._data['Interconnection_branches']:
            line[3] = aux_num_lines
            aux_num_lines += 1
        
        aux_num_lines = 0
        for line in copy_lines:
            self._data['Preserved_branches'].append([\
                line.get_element(name='position'), \
                line.get_element(name='bus_number')[0], \
                line.get_element(name='bus_number')[1],
                aux_num_lines, \
                line.get_element(name='bus_number')[0], \
                line.get_element(name='bus_number')[1]])
            aux_num_lines += 1
        
        self._data['Preserved_branches'].extend(self._data['Interconnection_branches'])

        copy_nodes.extend(new_nodes)
        copy_lines.extend(new_lines)
        self.set_objects(name='bus', list_obj=copy_nodes)
        self.set_objects(name='transmissionline', list_obj=copy_lines)
        self.update_all_positions()

    def __add_artificial_node(self, supernodes=None, counter=None, \
        number_node=None, new_nodes=None, pos_supernode=None):        
        ''' This class method add one artificial bus to the list of artificial 
        buses '''
        new_nodes[counter].set_element(name='voltage_kv', val=0)
        new_nodes[counter].set_element(name='name', \
            val=supernodes[pos_supernode]['coupling_bus_name'])
        new_nodes[counter].set_element(name='number', val=number_node)
        new_nodes[counter].set_element(name='type', val=1)
        new_nodes[counter].set_element(\
            name='maximum_voltage_magnitude', val=10)
        new_nodes[counter].set_element(\
            name='minimum_voltage_magnitude', val=0)
        new_nodes[counter].set_element(name='non_technical_losses_fix',\
            val=supernodes[pos_supernode][\
            'equivalent_non_technical_losses_fix'])
        new_nodes[counter].set_element(name='active_power_demand_peak',\
            val=supernodes[pos_supernode]['equivalent_active_power_demand'])
        new_nodes[counter].set_element(\
            name='reactive_power_demand_peak',\
            val=supernodes[pos_supernode]['equivalent_reactive_power_demand'])
        new_nodes[counter].set_element(\
            name='initial_voltage_magnitude', val=1)
        new_nodes[counter].set_element(\
            name='initial_voltage_angle', val=0)
        new_nodes[counter].set_element(\
            name='zone', val=supernodes[pos_supernode]['zone_supernode'])        
        return new_nodes

    def __add_artificial_lines(self, supernodes=None, counter=None, \
        number_line=None, new_nodes=None, counternode=None, number_node=None,
        pos_supernode=None, new_lines=None):
        ''' This class method adds the artificial transmission lines to the 
        electricity network '''
        xcounter = 0
        number_lines = []
        for xnodespos in supernodes[pos_supernode]['coupling_buses']:
            new_lines.append(TransmissionLine())
            new_lines[counter].set_element(name='shunt_susceptance', val=\
                supernodes[pos_supernode]['artificial_lines_info']\
                    [xcounter].get_element(name='shunt_susceptance'))
            new_lines[counter].set_element(name='resistance', val=\
                supernodes[pos_supernode]['artificial_lines_info']\
                    [xcounter].get_element(name='resistance'))
            new_lines[counter].set_element(name='reactance', val=\
                supernodes[pos_supernode]['artificial_lines_info']\
                    [xcounter].get_element(name='reactance'))
            new_lines[counter].set_element(name='number', val=number_line)
            new_lines[counter].set_element(name='long_term_thermal_limit', \
                val=supernodes[pos_supernode]['artificial_lines_info']\
                    [xcounter].get_element(name='long_term_thermal_limit'))
            new_lines[counter].set_element(name='non_technical_losses_fix', \
                val=supernodes[pos_supernode]['artificial_lines_info']\
                    [xcounter].get_element(name='non_technical_losses_fix'))
            new_lines[counter].set_element(name='bus_number', \
                val=[self._data['SupernodesNetworkInfo'][pos_supernode]\
                    ['coupling_buses'][xcounter].get_element(name='number'), \
                    number_node])
            new_lines[counter].set_element(name='status', \
                val=1)
            number_lines.append(number_line)
            self._data['Interconnection_branches'].append([\
                supernodes[pos_supernode]\
                ['artificial_lines_info'][xcounter].get_element(\
                name='position'), \
                supernodes[pos_supernode]\
                ['artificial_lines_info'][xcounter].get_element(\
                name='bus_number')[0], supernodes[pos_supernode]\
                ['artificial_lines_info'][xcounter].get_element(\
                name='bus_number')[1], -1,
                self._data['SupernodesNetworkInfo'][pos_supernode]\
                ['coupling_buses'][xcounter].get_element(name='number'), \
                number_node])
            xcounter += 1
            counter += 1
            number_line += 1
        new_nodes[counternode].set_element(name='transmissionline_number', \
            val=number_lines)

        return counter, number_line

    def __add_generators_to_artificial_node(self, supernodes=None, counter=None, \
        number_node=None, new_nodes=None, pos_supernode=None):
        ''' This class method adds generators to the artificial node '''
        for xgentype in self.get_generation_types_names():
            auxgennumber = []
            for xgen in self.get_objects(name=xgentype):
                if xgen.get_element(name='bus_position') in \
                    supernodes[pos_supernode]['bus']:
                    xgen.set_element(name='bus_number', val=number_node)
                    auxgennumber.append(number_node)
            new_nodes[counter].set_element(\
                name=xgentype+'_number', val=auxgennumber)

    def __check_if_empty_list(self, name=None, bus_position=None):
        ''' This method checks if the bus object is connected to the network '''
        assert bus_position is not None, "No bus position passed to check its \
            connectivity"
        assert name is not None, "No name of branch element pass"
        aux = self.get_object_elements(name_object='bus', \
            name_element=name+'_position', pos_object=bus_position)
        if aux != []:
            return False
    
    def __check_positions_branch_objects(self, name=None, bus_position=None, \
        vol_kv=None, flags=None, supernode=None, flag_supernode=None):
        assert bus_position is not None, "No bus position passed to reduce \
            the network"
        assert flags is not None, "No flag dictionary passed to reduce the \
            network"
        assert supernode is not None, "No supernode dictionary passed to \
            reduce the network"
        assert flag_supernode is not None, "No flag for the supernode passed to \
            reduce the network"
        assert name is not None, "No name of branch element pass"

        # Checking if the object exist
        aux = self.get_object_elements(name_object=name, \
            name_element='position', pos_object=0)
        if aux == None:
            return flags, supernode, flag_supernode
        
        auxlist = self.get_object_elements(name_object='bus', \
            name_element=name+'_position', pos_object=bus_position)
        if auxlist == None:
            return flags, supernode, flag_supernode

        for xauxlist in auxlist:
            if not flags[name][xauxlist]:
                list_pos_buses = self.get_object_elements(\
                    name_object=name, \
                    name_element='bus_position', pos_object=xauxlist)
                assert bus_position in list_pos_buses , \
                    "The node should be on the list list_pos_buses"
                flags, supernode, flag_supernode = self.__check_ends_lines(\
                    bus_position=bus_position, vol_kv=vol_kv, \
                    element_position=xauxlist, element_name=name, \
                    list_pos_buses=list_pos_buses, flags=flags, \
                    supernode=supernode, flag_supernode=flag_supernode)                
                if not flags[name][xauxlist]:
                    flags[name][xauxlist] = True
                    supernode[name].append(xauxlist)
        return flags, supernode, flag_supernode

    def __check_ends_lines(self, bus_position=None, \
        vol_kv=None, element_position=None, element_name=None, \
        list_pos_buses=None, flags=None, supernode=None, flag_supernode=None):
        ''' This class method check the nodes that are not connected to a three 
        winding transformer'''
        for xauxlisbuses in list_pos_buses:
            auxvoltage = self.get_object_elements(name_object='bus', \
                name_element='voltage_kv', pos_object=xauxlisbuses)
            if xauxlisbuses != bus_position and \
                not flags['bus'][xauxlisbuses]:
                flags, supernode, flag_supernode = \
                    self.__voltage_track(bus_position=xauxlisbuses, \
                    vol_kv=vol_kv, flags=flags, supernode=supernode, \
                    flag_supernode=flag_supernode)
                if  auxvoltage >= vol_kv and auxvoltage != \
                    self.get_element(name='voltagethreewindingtrafos'):
                    supernode['artificial_lines_info'].append(\
                        self.get_objects(name=element_name, pos=\
                            element_position))
            elif xauxlisbuses != bus_position and \
                flags['bus'][xauxlisbuses] and \
                auxvoltage >= vol_kv and \
                xauxlisbuses not in supernode['bus'] \
                and auxvoltage != self.get_element(\
                name='voltagethreewindingtrafos'):
                    supernode['coupling_buses'].append(xauxlisbuses)
                    supernode['artificial_lines_info'].append(\
                        self.get_objects(name=element_name, pos=\
                            element_position))
        return flags, supernode, flag_supernode

    def __find_voltage_levels(self):
        ''' This method finds all voltage levels in the system '''
        aux = []
        for xaux in range(self.get_no_objects(name='bus')):
            aux.append(self.get_object_elements(name_object='bus', \
                name_element='voltage_kv', pos_object=xaux))
        aux1 = []
        for xn in aux:
            flag_vol = False
            for xvol in aux1:
                if xn == xvol:
                    flag_vol = True
                    break
            if not flag_vol:
                aux1.append(xn)
        aux1.sort(reverse=True)
        self._data['VoltageLevels'] = aux1
    
    def __list_elements_to_remove(self, vol_kv=None, electricity_network=None):
        ''' This function return the list of elements (nodes, lines, trafos, 
        etc) to be eliminated in the network reduction'''
        copy_electricity_network = ElectricityNetwork()
        copy_electricity_network.set_electricity_network_data(\
            electricity_network)
    
    def __reduce_network_from_supernodes(self, supernodes=None):
        ''' This class method reduce the network based on the information of 
            supernodes '''
        assert isinstance(supernodes, list), "Incorrect \
            object passed for supernodes"
        
        self.__save_network_info_in_supernodes(supernodes=supernodes)
        busestoerase = []
        for xsuper in supernodes:
            xsuper['zone_supernode'] = self.get_object_elements(\
                    name_object='bus', \
                    name_element='zone', pos_object=\
                    xsuper['coupling_buses'][0])
            busestoerase.extend(xsuper['bus'])
            xsuper['equivalent_active_power_demand'] = 0
            xsuper['equivalent_reactive_power_demand'] = 0
            xsuper['equivalent_non_technical_losses_fix'] = 0
            for xbus in xsuper['bus']:
                xsuper['equivalent_active_power_demand'] += \
                    self.get_object_elements(\
                    name_object='bus', \
                    name_element='active_power_demand_peak', pos_object=xbus)
                xsuper['equivalent_reactive_power_demand'] += \
                    self.get_object_elements(\
                    name_object='bus', \
                    name_element='reactive_power_demand_peak', pos_object=xbus)
                xsuper['equivalent_non_technical_losses_fix'] += \
                    self.get_object_elements(\
                    name_object='bus', \
                    name_element='non_technical_losses_fix', pos_object=xbus)
        self.delete_objects(name='bus', pos=busestoerase)
        for xseries in self.get_series_elements_names():
            elementstoerase = []
            for xsuper in supernodes:
                elementstoerase.extend(xsuper[xseries])
            for ete in elementstoerase:
                bus_num = self.get_object_elements(\
                    name_object=xseries, \
                    name_element='bus_number', \
                    pos_object=ete)
                self._data['Removed_branches'].append(\
                    [bus_num[0], bus_num[1]])
        for xseries in self.get_series_elements_names():
            elementstoerase = []
            for xsuper in supernodes:
                elementstoerase.extend(xsuper[xseries])
                for xserieselements in xsuper[xseries]:
                    auxlossfix = self.get_object_elements(\
                        name_object=xseries, \
                        name_element='non_technical_losses_fix', \
                        pos_object=xserieselements)
                    if auxlossfix != None:
                        xsuper['equivalent_non_technical_losses_fix'] += \
                            auxlossfix
            self.delete_objects(name=xseries, pos=elementstoerase)

    def __save_network_info_in_supernodes(self, supernodes=None):
        ''' This method class save the network information for the supernodes 
        '''
        self._data['SupernodesNetworkInfo'] = []
        for xsuper in supernodes:
            supernode = { # This dictionary contains the positions of the elements
                # of the network that will be converted into a supernode
                'bus' : self.get_objects(name='bus', pos=xsuper['bus']),
                'coupling_buses' : self.get_objects(name='bus', \
                    pos=xsuper['coupling_buses']),
                'coupling_bus_name' : xsuper['coupling_bus_name'], # name of the coupling node  
                    # that interconnect the supernode with the reduced network
                'transmissionline' : self.get_objects(name='transmissionline', \
                    pos=xsuper['transmissionline']),
                'transformers' : self.get_objects(name='transformers', \
                    pos=xsuper['transformers'])
            }
            self._data['SupernodesNetworkInfo'].append(supernode)

    def __voltage_track(self, bus_position=None, vol_kv=None, flags=None, \
        supernode=None, flag_supernode=None):
        ''' Recursive function for finding next voltage level '''
        assert bus_position is not None, "No bus position passed to reduce \
            the network"
        assert flags is not None, "No flag dictionary passed to reduce the \
            network"
        assert supernode is not None, "No supernode dictionary passed to \
            reduce the network"
        assert flag_supernode is not None, "No flag for the supernode passed to \
            reduce the network"
        
        if not flag_supernode:
            aux = self.get_object_elements(name_object='bus', \
                name_element='voltage_kv', pos_object=bus_position)
            if aux >= vol_kv and aux != self.get_element(\
                name='voltagethreewindingtrafos'):
                flags['bus'][bus_position] = True
                return flags, supernode, flag_supernode
        else:
            aux = self.get_object_elements(name_object='bus', \
                name_element='voltage_kv', pos_object=bus_position)
            if aux >= vol_kv and aux != self.get_element(\
                name='voltagethreewindingtrafos'):
                # If a supernode exist add the node to the list of coupling 
                # nodes
                flags['bus'][bus_position] = True
                supernode['coupling_buses'].append(bus_position)
                return flags, supernode, flag_supernode
        
        if flags['bus'][bus_position]:
            return flags, supernode, flag_supernode
        
        flags['bus'][bus_position] = True
        is_disconnected = True
        for key in flags.keys():
            if key != 'bus':
                is_disconnected = self.__check_if_empty_list(\
                    name=key, bus_position=bus_position)
            if not is_disconnected:
                break
        if is_disconnected:
            return flags, supernode, flag_supernode
        
        flag_supernode = True
        supernode['bus'].append(bus_position)
        
        for xlist in self.get_series_elements_names():
            flags, supernode, flag_supernode = \
                self.__check_positions_branch_objects(name=xlist, \
                bus_position=bus_position, vol_kv=vol_kv, \
                flags=flags, supernode=supernode, flag_supernode=flag_supernode)
        
        return flags, supernode, flag_supernode
    
    def get_removed_lines(self):
        ''' This definition returns the list of lines that were removed from the system '''
        return self._data['Removed_branches']

class TemporalTree():
    
    G = nx.DiGraph()